from .normalize import _normalize, set_spacy, wiktionary, _get_spacy
from .normalize import is_in_dictionary as _is_in_dictionary
from .ocr_builder import get_wv_normalization

class W2VException(Exception):
	pass

def normalize_words(words, n_best=10, dictionary=None, all_candidates=True, correct_spelling_cache=True, return_scores=False):
	return _normalize(words, "normalization.pt", n_best=n_best, dictionary=dictionary, all_candidates=all_candidates,correct_spelling_cache=correct_spelling_cache, return_scores=return_scores)

def ocr_correct_words(words, n_best=10, dictionary=None, all_candidates=True, hybrid=False, hybrid_w2v_model=None,correct_spelling_cache=True, return_scores=False):
	if hybrid is True and hybrid_w2v_model is None:
		raise W2VException("W2V model not specified")
	if hybrid and return_scores:
		raise Exception("hybrid mode does not support scores")
	norms = _normalize(words, "ocr.pt", n_best=n_best, dictionary=dictionary, all_candidates=all_candidates,correct_spelling_cache=correct_spelling_cache,return_scores=return_scores)
	if hybrid:
		for i, l in enumerate(norms):
			if len(l) == 0:
				w2v_norm = get_wv_normalization(words[i], hybrid_w2v_model, dictionary, cache=correct_spelling_cache)
				if len(w2v_norm) > 0:
					l.append(w2v_norm)
		return norms
	else:
		return norms

def is_correctly_spelled(word, dictionary=None, cache=True):
	return _is_in_dictionary(word.lower(), dictionary, _get_spacy(), cache)



import requests
from bs4 import BeautifulSoup
from mikatools import *


url = "https://en.wiktionary.org/w/index.php?title=Category:English_lemmas&from=A"
base_url = "https://en.wiktionary.org/"

def get_pages(url):
	all_lemmas = []
	while True:
		r = requests.get(url)
		soup = BeautifulSoup(r.text, 'html.parser')
		base = soup.find("div", id="mw-pages")
		lemmas = base.find_all("li")
		for lemma in lemmas:
			text = lemma.get_text()
			all_lemmas.append(text)
		next_button = base.find("a", string="next page")
		if next_button is None:
			break
		url = base_url + next_button.get("href")
	return all_lemmas




from mikatools import script_path, json_load
from onmt.translate.translator import Translator
from onmt.model_builder import load_test_model
from onmt.translate import GNMTGlobalScorer
from itertools import islice, repeat
import configargparse as cfargparse
import spacy
import os


wiktionary = None

def load_wiktionary():
	global wiktionary
	if wiktionary is not None:
		return
	try:
		wiktionary = set([x.lower() for x in json_load(script_path("wiktionary_lemmas.json"))])
	except:
		print("run python -m natas.download")
		wiktionary = []

is_in_data_cache = {"ceec_eng":{}, "ocr_fin":{}}

def set_spacy(nlp):
	models["spacy"] = nlp

def _get_spacy():
	if "spacy" not in models:
		try:
			models["spacy"] = spacy.load('en_core_web_md')
		except IOError:
			raise Exception("Spacy model was not loaded! Run: python -m spacy download en_core_web_md")
	return models["spacy"]

def split_corpus(f, shard_size):
    if shard_size <= 0:
        yield f
    else:
        while True:
            shard = list(islice(f, shard_size))
            if not shard:
                break
            yield shard

models = {}

def is_in_dictionary(word, correct_lemmas, spacy_nlp, cache=True, cache_name="ceec_eng", lemmatize=True):
	if correct_lemmas is None:
		load_wiktionary()
		correct_lemmas = wiktionary
	if cache and word in is_in_data_cache:
		return is_in_data_cache["ceec_eng"][word]
	if word in correct_lemmas:
		is_in_data_cache["ceec_eng"][word] = True
		return True
	if not lemmatize:
		is_in_data_cache["ceec_eng"][word] = False
		return False
	try:
		res = spacy_nlp(word)
		if len(res) > 1:
			is_in_data_cache["ceec_eng"][word] = False
			return False
		lemma = res[0].lemma_
	except:
		is_in_data_cache["ceec_eng"][word] = False
		return False
	if lemma in correct_lemmas:
		is_in_data_cache["ceec_eng"][word] = True
		return True
	is_in_data_cache["ceec_eng"][word] = False
	return False

def _dict_filter(results, dictionary, all_candidates=True, correct_spelling_cache=True):
	output = []
	nlp = _get_spacy()

	for word in results:
		cor = []
		for candidate in word:
			score = None
			if isinstance(candidate, tuple):
				score = candidate[1]
				candidate = candidate[0]
			candidate = candidate.replace(" ", "")
			if is_in_dictionary(candidate, dictionary, nlp, correct_spelling_cache):
				if score is not None:
					cor.append([candidate, score])
				else:
					cor.append(candidate)
				if not all_candidates:
					break
		output.append(cor)
	return output

class opennmt_opts(cfargparse.ArgumentParser):
	"""docstring for opennmt_opts"""
	def __init__(self, model, **kwargs):
		self.models = [model]
		for k, v in kwargs.items():
			setattr(self,k,v)

class fake_stream(object):
	"""docstring for fake_stream"""
	def __init__(self):
		self.lines = []
		self.scores = []

	def write(self, line):
		self.lines.append(line)

	def close(self):
		pass
	def flush(self):
		pass

	def __repr__(self):
		return "\n".join(self.lines)

	def load_scores(self, translated):
		self.scores = [[score.item() for score in score_set] for score_set in translated[0]]

	def get_lines(self):
		return str(self).split("\n")

	def get_scores(self):
		return self.scores

def _chunks(l, n):
    n = max(1, n)
    return [l[i:i+n] for i in range(0, len(l), n)]

def _parse_fake_stream(stream, n_best=10):
	parts = stream.get_lines()
	k = n_best + 1
	del parts[k-1::k]
	return _chunks(parts, n_best)

def _default_kwargs(words=None,n_best=10):
	return {"tgt_prefix":False,"fp32":False,"int8":False, "ban_unk_token":False, "random_sampling_topp":0.0,"avg_raw_probs":False,"batch_size":30,"attn_debug":False,"src":words,"tgt":None,"alpha":0.0,"beta":-0.0,"length_penalty":"none","coverage_penalty":"none","gpu":-1,"n_best":n_best,"min_length":0,"max_length":100,"ratio":-0.0,"beam_size":5,"random_sampling_topk":0,"random_sampling_temp":1.0,"stepwise_penalty":False,"dump_beam":"","block_ngram_repeat":0,"ignore_when_blocking":[],"replace_unk":True,"phrase_table":"","data_type":"text","verbose":False,"report_bleu":False,"report_rouge":False,"report_time":False,"seed":829,"shard_size":0}


def _load_model(name):
	model_path = script_path("models/" + name)
	if not os.path.exists(model_path):
		model_path = name
	opt = opennmt_opts(model_path, **_default_kwargs())
	m = load_test_model(opt)
	models[name] = m

def _give_model(name):
	if name not in models:
		_load_model(name)
	return models[name]

def _split_words(words):
	return [" ".join(x.lower()) for x in words]

def call_onmt(words, model_name, n_best=10, return_scores=False):
	#Adapted code from OpenNMT translate.py

	stream = fake_stream()
	fields, model, model_opt = _give_model(model_name)
	opt = opennmt_opts("", **_default_kwargs(words, n_best))
	scorer = GNMTGlobalScorer.from_opt(opt)
	t = Translator.from_opt(model, fields, opt, model_opt, global_scorer=scorer, out_file=stream, report_score=False)
	src_shards = split_corpus(opt.src, opt.shard_size)
	tgt_shards = split_corpus(opt.tgt, opt.shard_size) \
		if opt.tgt is not None else repeat(None)
	shard_pairs = zip(src_shards, tgt_shards)
	for i, (src_shard, tgt_shard) in enumerate(shard_pairs):
		translated = t.translate(
						src=src_shard,
						tgt=tgt_shard,
						#src_dir=opt.batch_size,
						batch_size=opt.batch_size,
						attn_debug=opt.attn_debug
						)
		stream.load_scores(translated)
	chunks = _parse_fake_stream(stream, n_best)
	scores = stream.get_scores()
	if return_scores is True:
		return [list(zip(c,s)) for c, s in zip(chunks, scores)]
	else:
		return chunks

def _normalize(words, model_name, n_best=5, dictionary=None, all_candidates=True, correct_spelling_cache=True, return_scores=False):
	res = call_onmt(_split_words(words), model_name, n_best=n_best, return_scores=return_scores)
	if dictionary is None:
		load_wiktionary()
		dictionary = wiktionary
	res = _dict_filter(res, dictionary, all_candidates=all_candidates, correct_spelling_cache=correct_spelling_cache)
	return res





from .normalize import is_in_dictionary, wiktionary, _get_spacy, load_wiktionary
import distance
from collections import defaultdict

load_wiktionary()

def get_min_distance(word, words):
	min = 2000000
	min_word = ""
	for w in words:
		d = distance.levenshtein(w, word)
		if d < min:
			min = d
			min_word = w
	return min_word, min

def extract_parallel(seed_words, model, dictionary=wiktionary, lemmatize=True, use_freq=True, word_len=0, min_frequency=1000, cache=False, cache_name="ocr"):
	res = defaultdict(dict)
	for word in seed_words:
		if use_freq and seed_words[word] < min_frequency:
			continue
		word = word.strip().lower()
		if len(word) < word_len:
			continue
		errors = get_ocr_error_dict(word, model, dictionary, lemmatize=lemmatize, cache=cache,cache_name=cache_name)
		for correct in errors:
			res[correct].update(errors[correct])
	return res

def get_wv_normalization(word, model, dictionary, lemmatize=True, cache=True, cache_name="ocr"):
	res = get_ocr_error_dict(word,model, dictionary, lemmatize=lemmatize, cache=cache, cache_name=cache_name)
	for key, value in res.iteritems():
		if word in value and value[word] < 4:
			return key
	return ""

def get_ocr_error_dict(word, model, dictionary, lemmatize=True, cache=True, cache_name="ocr"):
	ocr_errors = []
	non_errors = [word]
	if lemmatize:
		spacy_nlp = _get_spacy()
	else:
		spacy_nlp = None
	try:
		pot_ocr_errors = model.most_similar(word)
	except:
		#word not in vocabulary
		return []

	for pot_ocr_error in pot_ocr_errors:
		pot_ocr_error = pot_ocr_error[0]
		if not is_in_dictionary(pot_ocr_error, dictionary, spacy_nlp, cache=cache, cache_name=cache_name,lemmatize=lemmatize):
			ocr_errors.append(pot_ocr_error)
		else:
			non_errors.append(pot_ocr_error)

	results = defaultdict(dict)
	for error in ocr_errors:
		w, d = get_min_distance(error, non_errors)
		results[w][error] = d
	return results




from mikatools import *
import os

def main():
	url = "https://github.com/mikahama/natas/raw/master/natas/models/"
	models = ["normalization_brnn_latech19.pt", "normalization.pt", "ocr_ranlp19.pt", "ocr.pt"]
	model_path = script_path("models")
	if not os.path.exists(model_path):
		os.mkdir(model_path)
	for i, model in enumerate(models):
		print("Downloading", i+1, "out of", len(models) )
		download_file(url + model, script_path("models/" + model), show_progress=True)

if __name__== "__main__":
  main()